﻿/*
 * Project Blend: the Blossom of .NET Development
 * Copyright (c) 2007 Project Blend
 * 
 * 이 라이브러리는 자유 소프트웨어입니다. 소프트웨어의 피양도자는 자유 소프트웨어 재단이
 * 공표한 GNU 약소 일반 공중 사용 허가서 2.1판 또는 그 이후 판을 임의로 선택해서, 그 규정에
 * 따라 라이브러리를 개작하거나 재배포할 수 있습니다.
 * 
 * 이 라이브러리는 유용하게 사용될 수 있으리라는 희망에서 배포되고 있지만, 특정한 목적에
 * 맞는 적합성 여부나 판매용으로 사용할 수 있으리라는 묵시적인 보증을 포함한 어떠한 형태의
 * 보증도 제공하지 않습니다. 보다 자세한 사항에 대해서는 GNU 약소 일반 공중 사용 허가서를
 * 참고하시기 바랍니다.
 * 
 * GNU 약소 일반 공중 사용 허가서는 이 라이브러리와 함께 제공됩니다. 만약, 이 문서가 누락
 * 되어있다면 자유 소프트웨어 재단으로 문의하시기 바랍니다.
 * 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * Korean Translation Comes from: http://korea.gnu.org/people/chsong/copyleft/lgpl.ko.html
 * */

using System;
using System.Text;
using System.Reflection;
using System.Reflection.Emit;
using System.Runtime.InteropServices;

namespace Blend.xPlatform
{
    [Serializable]
    public delegate Type TypeEvaluator(Type declaringType, Type originalType);

    /// <summary>
    /// Reflection Emitting API를 이용하여 컴파일 이후의 런타임 형식들을 제어하고 캐시합니다.
    /// </summary>
    public class DynamicTypeServices
    {
        /// <summary>
        /// 동적 어셈블리를 관리할 응용프로그램 도메인에 대한 참조입니다.
        /// </summary>
        private AppDomain targetAppDomain;

        /// <summary>
        /// 다수의 동적 모듈을 포함하는 동적 어셈블리에 대한 참조입니다.
        /// </summary>
        private AssemblyBuilder currentAssembly;

        /// <summary>
        /// 다수의 형식을 포함하는 동적 모듈에 대한 참조입니다.
        /// </summary>
        private ModuleBuilder currentModule;

        /// <summary>
        /// 지정된 응용프로그램 도메인에 새로운 동적 어셈블리와 동적 모듈을 구축합니다.
        /// </summary>
        /// <param name="targetAppDomain">
        /// 동적 어셈블리와 동적 모듈을 구축할 응용프로그램 도메인의 참조를
        /// 지정합니다. Null 참조를 지정하면 기본 응용프로그램 도메인이
        /// 자동으로 선택됩니다.</param>
        public DynamicTypeServices(AppDomain targetAppDomain)
        {
            this.targetAppDomain = targetAppDomain ?? AppDomain.CurrentDomain;
            
            this.currentAssembly = this.targetAppDomain.DefineDynamicAssembly(
                new AssemblyName(Guid.NewGuid().ToString("N")),
                AssemblyBuilderAccess.Run);

            this.currentModule = this.currentAssembly.DefineDynamicModule(
                Guid.NewGuid().ToString("N"), true);
        }

        /// <summary>
        /// 기본 응용프로그램 도메인에 새로운 동적 어셈블리와 동적 모듈을 구축합니다.
        /// </summary>
        public DynamicTypeServices()
            : this(null)
        {
        }

        /// <summary>
        /// 동적 어셈블리가 위치한 응용프로그램 도메인의 참조를 가져옵니다.
        /// </summary>
        public AppDomain AppDomain
        {
            get { return this.targetAppDomain; }
        }

        /// <summary>
        /// 동적 모듈이 위치한 동적 어셈블리의 참조를 가져옵니다.
        /// </summary>
        public Assembly Assembly
        {
            get { return this.currentAssembly; }
        }

        /// <summary>
        /// 동적 형식들이 위치한 동적 모듈의 참조를 가져옵니다.
        /// </summary>
        public Module Module
        {
            get { return this.currentModule; }
        }

        /// <summary>
        /// 변환하기 이전의 원본 형식과 변환 작업을 설명하는 GUID를 지정하여 동적 모듈 내에 캐쉬된 형식인지 확인합니다.
        /// </summary>
        /// <param name="guid">변환 작업을 설명하는 GUID입니다.</param>
        /// <param name="originalType">변환하기 이전의 원본 형식을 지정합니다.</param>
        /// <returns>동적 모듈 내에 생성된 형식이면 True, 그렇지 않으면 False를 반환합니다.</returns>
        public bool IsCachedType(Guid guid, Type originalType)
        {
            return (this.GetCachedType(guid, originalType) != null);
        }

        /// <summary>
        /// 변환하기 이전의 원본 형식과 변환 작업을 설명하는 GUID를 지정하여 편집 가능한 형식인지 확인합니다.
        /// </summary>
        /// <param name="guid">변환 작업을 설명하는 GUID입니다.</param>
        /// <param name="originalType">변환하기 이전의 원본 형식을 지정합니다.</param>
        /// <returns>동적 모듈 내에 생성된 형식이고 편집 가능한 상태이면 True, 그렇지 않으면 False를 반환합니다.</returns>
        public bool IsEditableType(Guid guid, Type originalType)
        {
            Type t = this.GetCachedType(guid, originalType);
            return (t != null && t is TypeBuilder);
        }

        /// <summary>
        /// 변환하기 이전의 원본 형식과 변환 작업을 설명하는 GUID를 지정하여 사용 가능한 형식인지 확인합니다.
        /// </summary>
        /// <param name="guid">변환 작업을 설명하는 GUID입니다.</param>
        /// <param name="originalType">변환하기 이전의 원본 형식을 지정합니다.</param>
        /// <returns>동적 모듈 내에 생성된 형식이고 사용 가능한 상태이면 True, 그렇지 않으면 False를 반환합니다.</returns>
        public bool IsFinalizedType(Guid guid, Type originalType)
        {
            Type t = this.GetCachedType(guid, originalType);
            return (t != null && !(t is TypeBuilder));
        }

        /// <summary>
        /// 변환하기 이전의 원본 형식과 변환 작업을 설명하는 GUID를 지정하여 동적 모듈 내의 형식을 가져옵니다.
        /// </summary>
        /// <param name="guid">변환 작업을 설명하는 GUID입니다.</param>
        /// <param name="originalType">변환하기 이전의 원본 형식을 지정합니다.</param>
        /// <returns>동적 모듈 내의 형식이 존재하면 형식 정보를, 그렇지 않으면 Null 참조를 반환합니다.</returns>
        public Type GetCachedType(Guid guid, Type originalType)
        {
            return this.currentModule.GetType(
                this.DecorateTypeName(guid, originalType));
        }

        public Type GetCachedType(Guid guid, string originalCustomName)
        {
            return this.currentModule.GetType(this.DecorateTypeName(guid, originalCustomName));
        }

        /// <summary>
        /// 변환하기 이전의 원본 형식과 변환 작업을 설명하는 GUID를 지정하여 동적 모듈 내에서 사용할 형식 이름을 생성합니다.
        /// </summary>
        /// <param name="guid">변환 작업을 설명하는 GUID입니다.</param>
        /// <param name="originalType">변환하기 이전의 원본 형식을 지정합니다.</param>
        /// <returns>동적 모듈 내에서 사용할 형식 이름이 반환됩니다.</returns>
        protected string DecorateTypeName(Guid guid, Type originalType)
        {
            return this.DecorateTypeName(guid, originalType.Namespace + "." + originalType.Name);
        }

        protected string DecorateTypeName(Guid guid, string customName)
        {
            string base64Format = Convert.ToBase64String(Encoding.Default.GetBytes(customName ?? Guid.NewGuid().ToString()), Base64FormattingOptions.None).Replace("+", "_P_").Replace("/", "_S_").Replace("=", "_E_").Replace(" ", "_W_");
            return (customName + "_CACHED_NAME_FOR_" + guid.ToString("N") + "_ORIGINAL_IS_" + base64Format);
        }

        /// <summary>
        /// 기본 형식 평가 루틴으로 관리 형식을 비 관리 형식으로 추천하여 변경합니다.
        /// </summary>
        /// <param name="declaringType">멤버를 선언하는 형식 (클래스, 구조체 등)입니다.</param>
        /// <param name="originalType">필드나 프로퍼티의 형식입니다.</param>
        /// <returns>추천하는 형식이 반환됩니다.</returns>
        protected virtual Type DefaultTypeEvaluator(Type declaringType, Type originalType)
        {
            if (originalType.Equals(typeof(string)))
                return typeof(IntPtr);

            return originalType;
        }

        /// <summary>
        /// 관리되는 형식의 필드를 하나 이상 포함하는 클래스나 구조체 형식을 순수한 구조체 형식으로 변환합니다.
        /// </summary>
        /// <remarks>
        /// finalize 인수를 True로 지정하면 멤버 구성을 변경할 수 없게 됩니다.
        /// </remarks>
        /// <param name="originalType">변환할 대상 형식을 지정합니다.</param>
        /// <param name="finalize">변환된 형식의 편집을 완료할 지의 여부입니다.</param>
        /// <param name="evaluator">형식 평가 및 변경을 위한 대리자를 지정합니다.</param>
        /// <returns>변환된 형식이 반환됩니다.</returns>
        public Type BuildStructureType(Type originalType, bool finalize, TypeEvaluator evaluator)
        {
            if (evaluator == null)
                evaluator = new TypeEvaluator(this.DefaultTypeEvaluator);

            Guid guid = new Guid("{46A32067-4B5C-4b36-A794-A55A74C18C17}");
            string newName = this.DecorateTypeName(guid, originalType);
            Type cachedType = this.GetCachedType(guid, originalType);

            if (cachedType != null)
                return cachedType;

            FieldInfo[] fieldList = originalType.GetFields();

            TypeBuilder conversionType = this.currentModule.DefineType(newName,
                TypeAttributes.Public | TypeAttributes.SequentialLayout | TypeAttributes.Serializable,
                typeof(ValueType));

            foreach (FieldInfo eachField in fieldList)
                conversionType.DefineField(
                    eachField.Name,
                    (evaluator(originalType, eachField.FieldType) ?? eachField.FieldType),
                    eachField.Attributes);

            return (finalize ? conversionType.CreateType() : conversionType);
        }

        /// <summary>
        /// 관리되는 형식의 필드를 하나 이상 포함하는 클래스나 구조체 형식을 순수한 구조체 형식으로 변환합니다.
        /// </summary>
        /// <param name="originalType">변환할 대상 형식을 지정합니다.</param>
        /// <returns>변환된 형식이 반환됩니다.</returns>
        public Type BuildStructureType(Type originalType)
        {
            return this.BuildStructureType(originalType, true, null);
        }

        public Type CreateDelegateType(string functionName, Type returnType, params Type[] fullParameterSpecs)
        {
            functionName = functionName ?? Guid.NewGuid().ToString("N");
            Guid guid = new Guid("{2E03FDD5-C88D-4ce3-B62A-143834C5DF9C}");
            string newName = this.DecorateTypeName(guid, functionName);
            Type cachedType = this.GetCachedType(guid, functionName);

            if (cachedType != null)
                return cachedType;

            TypeBuilder typeBuilder = this.currentModule.DefineType(
                newName,
                TypeAttributes.AutoClass | TypeAttributes.Sealed | TypeAttributes.Public,
                typeof(MulticastDelegate),
                new Type[0]);

            typeBuilder.DefineConstructor(
                MethodAttributes.RTSpecialName | MethodAttributes.HideBySig | MethodAttributes.Public, CallingConventions.Standard,
                new Type[] { typeof(object), typeof(IntPtr) }).SetImplementationFlags(MethodImplAttributes.CodeTypeMask);

            typeBuilder.DefineMethod("Invoke",
                MethodAttributes.VtableLayoutMask | MethodAttributes.HideBySig | MethodAttributes.Virtual | MethodAttributes.Public, returnType, fullParameterSpecs).SetImplementationFlags(MethodImplAttributes.CodeTypeMask);

            return (cachedType = typeBuilder.CreateType());
        }

        public Delegate CreateCustomDelegate(IntPtr nativeFuntionPointer, string functionName, Type returnType, params Type[] fullParameterSpecs)
        {
            return CreateCustomDelegate(
                nativeFuntionPointer,
                this.CreateDelegateType(functionName, returnType, fullParameterSpecs));
        }

        public Delegate CreateCustomDelegate(IntPtr nativeFunctionPointer, Type delegateType)
        {
            return Marshal.GetDelegateForFunctionPointer(
                nativeFunctionPointer, delegateType);
        }
    }
}
